unit GsvListBox;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, Forms;

const
  WM_GSVLISTBOX_INVALIDATE = WM_USER + $2000;

type
  TGsvListBoxDrawItem = procedure(Sender: TObject; aCanvas: TCanvas;
                        const aRect: TRect; aIndex: Integer) of object;

  TGsvCustomListBox = class(TCustomControl)
  private
    FBorderStyle:       TBorderStyle;
    FHideSelection:     Boolean;
    FMinItemHeight:     Integer;
    FCount:             Integer;
    FItemIndex:         Integer;
    FTopIndex:          Integer;
    FVisibleCount:      Integer; //    
    FItemHeight:        Integer;

    FOnDrawItem:        TGsvListBoxDrawItem;
    FOnChangeSelection: TNotifyEvent;

    function  CorrectItemIndex(Value: Integer): Integer;
    function  CorrectTopIndex(Value: Integer): Integer;
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetCount(Value: Integer);
    procedure SetItemIndex(Value: Integer);
    procedure SetTopIndex(Value: Integer);
    procedure SetMinItemHeight(Value: Integer);
    procedure OnFontChanged(Sender: TObject);
    procedure SetItemHeight;
    procedure AdjustParams;

    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure CMCtl3DChanged(var Msg: TMessage); message CM_CTL3DCHANGED;
    procedure WMSetFocus(var Msg: TMessage); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TMessage); message WM_KILLFOCUS;
    procedure WMGsvListBoxInvalidate(var Msg: TMessage); message WM_GSVLISTBOX_INVALIDATE;

  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Resize; override;
    procedure Paint; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
              X, Y: Integer); override;
    function  DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function  DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;

    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property HideSelection: Boolean read FHideSelection write FHideSelection default True;
    property MinItemHeight: Integer read FMinItemHeight
             write SetMinItemHeight default 0;
    property OnDrawItem: TGsvListBoxDrawItem read FOnDrawItem write FOnDrawItem;
    property OnChangeSelection: TNotifyEvent read FOnChangeSelection
             write FOnChangeSelection;

  public
    constructor Create(aOwner: TComponent); override;

    procedure SetItemIndexCentered(Value: Integer);
    function  HitTest(Y: Integer): Integer;

    property Count: Integer read FCount write SetCount;
    property ItemIndex: Integer read FItemIndex write SetItemIndex;
    property TopIndex: Integer read FTopIndex write SetTopIndex;
    property VisibleCount: Integer read FVisibleCount;
  end;

  TGsvListBox = class(TGsvCustomListBox)
  published
    property Align;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Constraints;
    property Enabled;
    property Color default clWindow;
    property Ctl3D;
    property OnDrawItem;
    property Font;
    property HideSelection;
    property MinItemHeight;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnCanResize;
    property OnChangeSelection;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Gsv', [TGsvListBox]);
end;

{ TGsvCustomListBox }

function TGsvCustomListBox.CorrectItemIndex(Value: Integer): Integer;
begin
  if Value < 0 then
    Result := 0
  else if Value > (FCount - 1) then
    Result := FCount - 1
  else
    Result := Value;
end;

function TGsvCustomListBox.CorrectTopIndex(Value: Integer): Integer;
begin
  if Value < 0 then
    Result := 0
  else if (Value + FVisibleCount) > FCount then begin
    Result := FCount - FVisibleCount;
    if Result < 0 then
      Result := 0;
  end
  else
    Result := Value;
end;

procedure TGsvCustomListBox.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TGsvCustomListBox.SetCount(Value: Integer);
begin
  if Value < 0 then
    Value := 0;
  if FCount <> Value then begin
    FCount := Value;
    AdjustParams;
  end;
end;

procedure TGsvCustomListBox.SetItemIndex(Value: Integer);
var
  ti:  Integer;
  inv: Boolean;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  ti := FTopIndex;
  Value := CorrectItemIndex(Value);
  if Value < ti then
    ti := Value
  else if Value >= (ti + FVisibleCount) then
    ti := Value - FVisibleCount + 1;
  ti := CorrectTopIndex(ti);
  inv := False;
  if FItemIndex <> Value then begin
    FItemIndex := Value;
    inv := True;
    if Assigned(FOnChangeSelection) then
      FOnChangeSelection(Self);
  end;
  if FTopIndex <> ti then begin
    FTopIndex  := ti;
    SetScrollPos(Handle, SB_VERT, FTopIndex, True);
    inv := True;
  end;
  if inv then
    Invalidate;
end;

procedure TGsvCustomListBox.SetTopIndex(Value: Integer);
begin
  if FItemHeight = 0 then
    SetItemHeight;
  Value := CorrectTopIndex(Value);
  if FTopIndex <> Value then begin
    FTopIndex := Value;
    SetScrollPos(Handle, SB_VERT, FTopIndex, True);
    Invalidate;
  end;
end;

procedure TGsvCustomListBox.SetMinItemHeight(Value: Integer);
begin
  if FMinItemHeight <> Value then begin
    FMinItemHeight := Value;
    FItemHeight    := 0;
    AdjustParams;
  end;
end;

procedure TGsvCustomListBox.OnFontChanged(Sender: TObject);
begin
  FItemHeight := 0;
  AdjustParams;
end;

procedure TGsvCustomListBox.SetItemHeight;
begin
  Canvas.Font := Font;
  FItemHeight := Canvas.TextHeight('M');
  if FItemHeight = 0 then
    FItemHeight := 16;
  if FItemHeight < FMinItemHeight then
    FItemHeight := FMinItemHeight;
  FVisibleCount := ClientHeight div FItemHeight;
end;

procedure TGsvCustomListBox.AdjustParams;
var
  si: TScrollInfo;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  if FCount = 0 then begin
    FItemIndex := -1;
    FTopIndex  := 0;
  end
  else begin
    FItemIndex := CorrectItemIndex(FItemIndex);
    FTopIndex  := CorrectTopIndex(FTopIndex);
  end;
  if FCount > FVisibleCount then begin
    ZeroMemory(@si, SizeOf(si));
    si.cbSize := SizeOf(TScrollInfo);
    si.fMask  := SIF_RANGE or SIF_POS;
    si.nMin   := 0;
    si.nMax   := FCount - FVisibleCount;
    si.nPos   := FTopIndex;
    SetScrollInfo(Handle, SB_VERT, si, True);
  end
  else
    ShowScrollBar(Handle, SB_VERT, False);
  Invalidate;
end;

procedure TGsvCustomListBox.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  inherited;
  Msg.Result := DLGC_WANTARROWS;
end;

procedure TGsvCustomListBox.WMVScroll(var Msg: TWMVScroll);
var
  si: TScrollInfo;
begin
  with Msg do begin
    case ScrollCode of
      SB_LINEUP:   SetTopIndex(FTopIndex - 1);
      SB_LINEDOWN: SetTopIndex(FTopIndex + 1);
      SB_PAGEUP:   SetTopIndex(FTopIndex - FVisibleCount);
      SB_PAGEDOWN: SetTopIndex(FTopIndex + FVisibleCount);
      SB_THUMBPOSITION,
      SB_THUMBTRACK: begin
        ZeroMemory(@si, SizeOf(si));
        si.cbSize := SizeOf(si);
        SI.fMask  := SIF_TRACKPOS;
        if GetScrollInfo(Handle, SB_VERT, si) then
          SetTopIndex(si.nTrackPos);
      end;
    end;
  end;
end;

procedure TGsvCustomListBox.CMCtl3DChanged(var Msg: TMessage);
begin
  if NewStyleControls then
    RecreateWnd;
  inherited;
end;

procedure TGsvCustomListBox.WMSetFocus(var Msg: TMessage);
begin
  inherited;
  if FHideSelection then
    PostMessage(Handle, WM_GSVLISTBOX_INVALIDATE, 0, 0);
end;

procedure TGsvCustomListBox.WMKillFocus(var Msg: TMessage);
begin
  inherited;
  if FHideSelection then
    PostMessage(Handle, WM_GSVLISTBOX_INVALIDATE, 0, 0);
end;

procedure TGsvCustomListBox.WMGsvListBoxInvalidate(var Msg: TMessage);
begin
  Invalidate;
end;

procedure TGsvCustomListBox.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do begin
    Style := Style or BorderStyles[FBorderStyle] or WS_VSCROLL;
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then begin
      Style   := Style and (not WS_BORDER);
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TGsvCustomListBox.Resize;
begin
  inherited;
  FItemHeight := 0;
  AdjustParams;
end;

procedure TGsvCustomListBox.Paint;
var
  r:     TRect;
  i, im: Integer;
begin
  if (FCount <> 0) and Assigned(FOnDrawItem) then begin
    if FItemHeight = 0 then
      SetItemHeight;
    r := GetClientRect;
    im := FTopIndex + FVisibleCount + 1;
    if im >= Pred(FCount) then
      im := Pred(FCount);
    SetBkMode(Canvas.Handle, TRANSPARENT);
    for i := FTopIndex to im do begin
      r.Bottom := r.Top + FItemHeight;
      if i = FItemIndex then begin
        if FHideSelection and (not Focused) then begin
          Canvas.Brush.Color := clBtnFace;
          Canvas.Font.Color  := clBtnText;
        end
        else begin
          Canvas.Brush.Color := clHighlight;
          Canvas.Font.Color  := clHighlightText;
        end;
        Canvas.FillRect(r);
        SetBkMode(Canvas.Handle, TRANSPARENT);
      end;
      FOnDrawItem(Self, Canvas, r, i);
      if i = FItemIndex then begin
        Canvas.Brush.Color := Color;
        Canvas.Font.Color  := Font.Color;
        SetBkMode(Canvas.Handle, TRANSPARENT);
      end;
      r.Top := r.Bottom;
    end;
  end;
end;

procedure TGsvCustomListBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  case Key of
    VK_UP:    SetItemIndex(FItemIndex - 1);
    VK_DOWN:  SetItemIndex(FItemIndex + 1);
    VK_PRIOR: SetItemIndex(FItemIndex - FVisibleCount);
    VK_NEXT:  SetItemIndex(FItemIndex + FVisibleCount);
    VK_HOME:  SetItemIndex(0);
    VK_END:   SetItemIndex(FCount);
  end;
end;

procedure TGsvCustomListBox.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if TabStop and (not Focused) then
    SetFocus;
  if FItemHeight = 0 then
    SetItemHeight;
  SetItemIndex(FTopIndex + (Y div FItemHeight));
end;

function TGsvCustomListBox.DoMouseWheelDown(Shift: TShiftState;
  MousePos: TPoint): Boolean;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
  SetTopIndex(FTopIndex + 1);
end;

function TGsvCustomListBox.DoMouseWheelUp(Shift: TShiftState;
  MousePos: TPoint): Boolean;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
  SetTopIndex(FTopIndex - 1);
end;

constructor TGsvCustomListBox.Create(AOwner: TComponent);
const
  st = [csCaptureMouse, csOpaque, csClickEvents, csDoubleClicks, csNeedsBorderPaint];
begin
  inherited Create(AOwner);
  if NewStyleControls then
    ControlStyle := st
  else
    ControlStyle := st + [csFramed];
  FBorderStyle   := bsSingle;
  FHideSelection := True;
  FItemIndex     := -1;
  ParentColor    := False;
  Color          := clWindow;
  DoubleBuffered := True;
  TabStop        := True;
  Font.OnChange  := OnFontChanged;
end;

procedure TGsvCustomListBox.SetItemIndexCentered(Value: Integer);
var
  ti:  Integer;
  inv: Boolean;
begin
  if FItemHeight = 0 then
    SetItemHeight;
  Value := CorrectItemIndex(Value);
  inv   := False;
  if (Value >= FTopIndex) and (Value < (FTopIndex + FVisibleCount)) then
    ti := FTopIndex
  else
    ti := CorrectTopIndex(Value - (FVisibleCount div 2));
  if FItemIndex <> Value then begin
    FItemIndex := Value;
    inv := True;
    if Assigned(FOnChangeSelection) then
      FOnChangeSelection(Self);
  end;
  if FTopIndex <> ti then begin
    FTopIndex  := ti;
    SetScrollPos(Handle, SB_VERT, FTopIndex, True);
    inv := True;
  end;
  if inv then
    Invalidate;
end;

function TGsvCustomListBox.HitTest(Y: Integer): Integer;
begin
  Result := FTopIndex + (Y div FItemHeight);
  if Result >= FCount then
    Result := -1;
end;

end.
